@@ -31,9 +31,11 @@ import (
 )
 
 type HardwareRandomObserveInterfaceSuite struct {
-	iface interfaces.Interface
-	slot  *interfaces.Slot
-	plug  *interfaces.Plug
+	iface    interfaces.Interface
+	slotInfo *snap.SlotInfo
+	slot     *interfaces.ConnectedSlot
+	plugInfo *snap.PlugInfo
+	plug     *interfaces.ConnectedPlug
 }
 
 var _ = Suite(&HardwareRandomObserveInterfaceSuite{
@@ -41,20 +43,22 @@ var _ = Suite(&HardwareRandomObserveInterfaceSuite{
 })
 
 const hardwareRandomObserveConsumerYaml = `name: consumer
+version: 0
 apps:
  app:
   plugs: [hardware-random-observe]
 `
 
 const hardwareRandomObserveCoreYaml = `name: core
+version: 0
 type: os
 slots:
   hardware-random-observe:
 `
 
 func (s *HardwareRandomObserveInterfaceSuite) SetUpTest(c *C) {
-	s.plug = MockPlug(c, hardwareRandomObserveConsumerYaml, nil, "hardware-random-observe")
-	s.slot = MockSlot(c, hardwareRandomObserveCoreYaml, nil, "hardware-random-observe")
+	s.plug, s.plugInfo = MockConnectedPlug(c, hardwareRandomObserveConsumerYaml, nil, "hardware-random-observe")
+	s.slot, s.slotInfo = MockConnectedSlot(c, hardwareRandomObserveCoreYaml, nil, "hardware-random-observe")
 }
 
 func (s *HardwareRandomObserveInterfaceSuite) TestName(c *C) {
@@ -62,34 +66,27 @@ func (s *HardwareRandomObserveInterfaceSuite) TestName(c *C) {
 }
 
 func (s *HardwareRandomObserveInterfaceSuite) TestSanitizeSlot(c *C) {
-	c.Assert(s.slot.Sanitize(s.iface), IsNil)
-	slot := &interfaces.Slot{SlotInfo: &snap.SlotInfo{
-		Snap:      &snap.Info{SuggestedName: "some-snap"},
-		Name:      "hardware-random-observe",
-		Interface: "hardware-random-observe",
-	}}
-	c.Assert(slot.Sanitize(s.iface), ErrorMatches,
-		"hardware-random-observe slots are reserved for the core snap")
+	c.Assert(interfaces.BeforePrepareSlot(s.iface, s.slotInfo), IsNil)
 }
 
 func (s *HardwareRandomObserveInterfaceSuite) TestSanitizePlug(c *C) {
-	c.Assert(s.plug.Sanitize(s.iface), IsNil)
+	c.Assert(interfaces.BeforePreparePlug(s.iface, s.plugInfo), IsNil)
 }
 
 func (s *HardwareRandomObserveInterfaceSuite) TestAppArmorSpec(c *C) {
 	spec := &apparmor.Specification{}
-	c.Assert(spec.AddConnectedPlug(s.iface, s.plug, nil, s.slot, nil), IsNil)
+	c.Assert(spec.AddConnectedPlug(s.iface, s.plug, s.slot), IsNil)
 	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
 	c.Assert(spec.SnippetForTag("snap.consumer.app"), testutil.Contains, "hw_random/rng_{available,current} r,")
 }
 
 func (s *HardwareRandomObserveInterfaceSuite) TestUDevSpec(c *C) {
 	spec := &udev.Specification{}
-	c.Assert(spec.AddConnectedPlug(s.iface, s.plug, nil, s.slot, nil), IsNil)
+	c.Assert(spec.AddConnectedPlug(s.iface, s.plug, s.slot), IsNil)
 	c.Assert(spec.Snippets(), HasLen, 2)
 	c.Assert(spec.Snippets(), testutil.Contains, `# hardware-random-observe
 KERNEL=="hwrng", TAG+="snap_consumer_app"`)
-	c.Assert(spec.Snippets(), testutil.Contains, `TAG=="snap_consumer_app", RUN+="/lib/udev/snappy-app-dev $env{ACTION} snap_consumer_app $devpath $major:$minor"`)
+	c.Assert(spec.Snippets(), testutil.Contains, `TAG=="snap_consumer_app", RUN+="/usr/lib/snapd/snap-device-helper $env{ACTION} snap_consumer_app $devpath $major:$minor"`)
 }
 
 func (s *HardwareRandomObserveInterfaceSuite) TestStaticInfo(c *C) {
@@ -101,7 +98,7 @@ func (s *HardwareRandomObserveInterfaceSuite) TestStaticInfo(c *C) {
 }
 
 func (s *HardwareRandomObserveInterfaceSuite) TestAutoConnect(c *C) {
-	c.Assert(s.iface.AutoConnect(s.plug, s.slot), Equals, true)
+	c.Assert(s.iface.AutoConnect(s.plugInfo, s.slotInfo), Equals, true)
 }
 
 func (s *HardwareRandomObserveInterfaceSuite) TestInterfaces(c *C) {
