@@ -29,7 +29,6 @@ const systemObserveBaseDeclarationSlots = `
     deny-auto-connection: true
 `
 
-// http://bazaar.launchpad.net/~ubuntu-security/ubuntu-core-security/trunk/view/head:/data/apparmor/policygroups/ubuntu-core/16.04/system-observe
 const systemObserveConnectedPlugAppArmor = `
 # Description: Can query system status information. This is restricted because
 # it gives privileged read access to all processes on the system and should
@@ -38,33 +37,42 @@ const systemObserveConnectedPlugAppArmor = `
 # Needed by 'ps'
 @{PROC}/tty/drivers r,
 
-# This ptrace is an information leak
+# This ptrace is an information leak. Intentionlly omit 'ptrace (trace)' here
+# since since ps doesn't actually need to trace other processes. Note this
+# allows a number of accesses (assuming the associated /proc file is allowed),
+# such as various memory address locations and esp/eip via /proc/*/stat,
+# /proc/*/mem, /proc/*/personality, /proc/*/stack, /proc/*/syscall,
+# /proc/*/timerslack_ns and /proc/*/wchan (see man proc).
+#
+# Some files like /proc/kallsyms (but anything using %pK format specifier) need
+# 'capability syslog' when /proc/sys/kernel/kptr_restrict=1, but we
+# intentionally do not allow since it could be used to defeat KASLR.
 ptrace (read),
 
-# ptrace can be used to break out of the seccomp sandbox, but ps requests
-# 'ptrace (trace)' even though it isn't tracing other processes. Unfortunately,
-# this is due to the kernel overloading trace such that the LSMs are unable to
-# distinguish between tracing other processes and other accesses. We deny the
-# trace here to silence the log.
-# Note: for now, explicitly deny to avoid confusion and accidentally giving
-# away this dangerous access frivolously. We may conditionally deny this in the
-# future.
-deny ptrace (trace),
-
 # Other miscellaneous accesses for observing the system
+@{PROC}/locks r,
+@{PROC}/modules r,
 @{PROC}/stat r,
 @{PROC}/vmstat r,
 @{PROC}/diskstats r,
 @{PROC}/kallsyms r,
+@{PROC}/partitions r,
+@{PROC}/sys/kernel/panic r,
+@{PROC}/sys/kernel/panic_on_oops r,
+@{PROC}/sys/vm/panic_on_oom r,
 
 # These are not process-specific (/proc/*/... and /proc/*/task/*/...)
 @{PROC}/*/{,task/,task/*/} r,
 @{PROC}/*/{,task/*/}auxv r,
+@{PROC}/*/{,task/*/}cgroup r,
 @{PROC}/*/{,task/*/}cmdline r,
+@{PROC}/*/{,task/*/}comm r,
 @{PROC}/*/{,task/*/}exe r,
+@{PROC}/*/{,task/*/}fdinfo/* r,
 @{PROC}/*/{,task/*/}stat r,
 @{PROC}/*/{,task/*/}statm r,
 @{PROC}/*/{,task/*/}status r,
+@{PROC}/*/{,task/*/}wchan r,
 
 # Allow discovering the os-release of the host
 /var/lib/snapd/hostfs/etc/os-release rk,
@@ -72,20 +80,20 @@ deny ptrace (trace),
 
 #include <abstractions/dbus-strict>
 
+# do not use peer=(label=unconfined) here since this is DBus activated
 dbus (send)
     bus=system
     path=/org/freedesktop/hostname1
     interface=org.freedesktop.DBus.Properties
-    member=Get{,All}
-    peer=(label=unconfined),
+    member=Get{,All},
 
 # Allow clients to introspect hostname1
+# do not use peer=(label=unconfined) here since this is DBus activated
 dbus (send)
     bus=system
     path=/org/freedesktop/hostname1
     interface=org.freedesktop.DBus.Introspectable
-    member=Introspect
-    peer=(label=unconfined),
+    member=Introspect,
 
 # Allow clients to enumerate DBus connection names on common buses
 dbus (send)
@@ -94,6 +102,14 @@ dbus (send)
     interface=org.freedesktop.DBus
     member=ListNames
     peer=(label=unconfined),
+
+# Allow clients to obtain the DBus machine ID on common buses. We do not
+# mediate the path since any peer can be used.
+dbus (send)
+    bus={session,system}
+    interface=org.freedesktop.DBus.Peer
+    member=GetMachineId
+    peer=(label=unconfined),
 `
 
 const systemObserveConnectedPlugSecComp = `
@@ -119,5 +135,6 @@ func init() {
 		connectedPlugAppArmor: systemObserveConnectedPlugAppArmor,
 		connectedPlugSecComp:  systemObserveConnectedPlugSecComp,
 		reservedForOS:         true,
+		suppressPtraceTrace:   true,
 	})
 }
